local standardFrames = {} -- Table to hold all standard unit frames
local compactFrames = {} -- Table to hold all compact unit frames

local frame = CreateFrame("FRAME"); -- Need a frame to respond to events
frame:RegisterEvent("ADDON_LOADED"); -- Fired when saved variables are loaded

function frame:OnEvent(event, arg1)
 if event == "ADDON_LOADED" and arg1 == "ShieldFrames" then
    standard = standard or 10
    compact = compact or 10
 end
end
frame:SetScript("OnEvent", frame.OnEvent);

SLASH_SHIELDFRAMES1 = "/shieldframes"
SLASH_SHIELDFRAMES2 = "/shfr"

function SlashCmdList.SHIELDFRAMES(msg)
   local command, rest = msg:match("^(%S*)%s*(.-)$");
   -- Any leading non-whitespace is captured into command;
   -- the rest (minus leading whitespace) is captured into rest.
   if command == "standard" then
      if tonumber(rest) ~= nil then
         standard = tonumber(rest)
         ShieldFrames_UpdateAll()
      else
         PrintHelp()
      end
   elseif command == "compact" then
      if tonumber(rest) ~=nil then
         compact = tonumber(rest)
         ShieldFrames_UpdateAll()
      end
   else
      PrintHelp()
   end
end

function ShieldFrames_UpdateAll()
   for key, value in pairs(standardFrames) do
      ShieldFrame_Update(value)
   end
   for key, value in pairs(compactFrames) do
      CompactShieldFrame_Update(value)
   end
end

function PrintHelp()
   print "\"/shieldframes standard <number>\" to change the number of health bar divisions for standard unit frames."
   print "\"/shieldframes compact <number>\" to change the number of health bar divisions for compact unit frames."
end

-- Hook the heal prediction functions so the addon functions run after them
-- Hook standard unit frames
hooksecurefunc("UnitFrame_SetUnit", function(self, ...)
				  if (self.shfr == nil) then
					table.insert(standardFrames, self)
					self.shfr = true
				  end
end)
hooksecurefunc("UnitFrameHealPredictionBars_Update", function(self, ...)
                  ShieldFrame_Update(self)
end)
-- Hook compact unit frames
hooksecurefunc("CompactUnitFrame_SetUnit", function(self, ...)
				  if (self.shfr == nil) then
					table.insert(compactFrames, self)
					self.shfr = true
				  end
end)
hooksecurefunc("CompactUnitFrame_UpdateHealPrediction", function(self, ...)
                    CompactShieldFrame_Update(self)
end)
hooksecurefunc("CompactUnitFrame_UpdateInRange", function(self, ...)
				  if (self.compactInitialize == nil) then
					self.compactInitialize = 0
				  end
				  if (self.compactInitialize < 2) then
					  local unit = self.unit or self.displayedUnit
					  SetHealthTickScale(self.healthBar, unit)
					  self.compactInitialize = self.compactInitialize + 1
				  end
end)

function ShieldFrame_Update (self)
   -- Sets up on first run, takes a unit frame as an argument
   local health = self.healthbar
   if (self.healthbar.healthTicks == nil) then
      -- If this is the first time running, create and set up the health tick texture
      health.healthTicks = {}
   end
   SetHealthTickScale(health, self.unit, 0)
   ShieldFrameHealthBar_Update(self.healthbar, self.unit);
end

function CompactShieldFrame_Update(self)
   -- Run the max health changing function
   local health = self.healthBar
   if self.healthBar.healthTicks == nil then
      health.healthTicks = {}
   end
   local unit = self.unit or self.displayedUnit
   SetHealthTickScale(health, unit, 1)
   CompactShieldFrame_UpdateMaxHealth(self)
end

function SetHealthTickScale(frame, unit, frameType)
   -- Changes the horizontal scaling of a health tick texture, takes a frame, unit, and frame type as arguments
   if (frame ~= nil and unit ~= nil) then
	   local scale = (UnitHealthMax(unit) + UnitGetTotalAbsorbs(unit)) / UnitHealthMax(unit)
	   local ShieldFrameTicksPerBar = 10
	   if frameType == 0 then
		  ShieldFrameTicksPerBar = standard
	   elseif frameType == 1 then
		  ShieldFrameTicksPerBar = compact
	   end
	   local width = frame:GetWidth()/((ShieldFrameTicksPerBar*scale))
	   local visible = ShieldFrameTicksPerBar*scale
	   local textureScale = 8+(ShieldFrameTicksPerBar/15)
	   local tickWidth = frame:GetWidth()/textureScale
	   local i = 1
	   while (frame.healthTicks[i] ~= nil) or (i <= visible) do
		  if frame.healthTicks[i] == nil then
			 frame.healthTicks[i] = frame:CreateTexture("HealthTick"..i, "ARTWORK", frame);
			 frame.healthTicks[i]:SetDrawLayer("ARTWORK", 7)
			 frame.healthTicks[i]:SetTexture("Interface/MainMenuBar/UI-ExhaustionTickNormal", false);
			 frame.healthTicks[i]:SetVertexColor(0, 0, 0)
		  end
		  if i > visible then
			 frame.healthTicks[i]:Hide()
		  else
			 frame.healthTicks[i]:Show()
		  end
		  frame.healthTicks[i]:SetTexCoord(0, textureScale, 0.5, 0.5)
		  frame.healthTicks[i]:SetPoint("TOPLEFT", frame, "TOPLEFT", (i*width) - tickWidth/2, 0)
		  frame.healthTicks[i]:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", (i*width) - tickWidth/2, 0)
		  i = i + 1
	   end
	end
end

function ShieldFrameHealthBar_Update(statusbar, unit)
   local totalAbsorb = UnitGetTotalAbsorbs(unit) or 0;
   if ( not statusbar or statusbar.lockValues ) then
      return;
   end
   
   if ( unit == statusbar.unit ) then
      local maxValue = UnitHealthMax(unit);
      
      -- Safety check to make sure we never get an empty bar.
      statusbar.forceHideText = false;
      if ( maxValue == 0 ) then
         maxValue = 1;
      end
      
      statusbar:SetMinMaxValues(0, (maxValue + totalAbsorb));
      ShieldStatusBar_UpdateTextString(statusbar, UnitHealthMax(unit));
      ShieldFrameHealPredictionBars_Update(statusbar:GetParent());
   end
end

function CompactShieldFrame_UpdateMaxHealth(frame)
   local unit = frame.unit or frame.displayedUnit
   local maxHealth = UnitHealthMax(unit);
   local totalAbsorb = UnitGetTotalAbsorbs(unit) or 0;
   frame.healthBar:SetMinMaxValues(0, (maxHealth + totalAbsorb));
   CompactShieldFrame_UpdateHealPrediction(frame);
end

local MAX_INCOMING_HEAL_OVERFLOW = 1.0;
function ShieldFrameHealPredictionBars_Update(frame)
   if ( not frame.myHealPredictionBar ) then
      return;
   end
   
   local maxHealth = UnitHealthMax(frame.unit);
   local health = frame.healthbar:GetValue();
   if ( maxHealth <= 0 ) then
      return;
   end
   
   local myIncomingHeal = UnitGetIncomingHeals(frame.unit, "player") or 0;
   local allIncomingHeal = UnitGetIncomingHeals(frame.unit) or 0;
   local totalAbsorb = UnitGetTotalAbsorbs(frame.unit) or 0;
   local textureScale = (maxHealth + totalAbsorb)/maxHealth;
   
   local myCurrentHealAbsorb = 0;
   if ( frame.healAbsorbBar ) then
      myCurrentHealAbsorb = UnitGetTotalHealAbsorbs(frame.unit) or 0;
      
      --We don't fill outside the health bar with healAbsorbs.  Instead, an overHealAbsorbGlow is shown.
      if ( health < myCurrentHealAbsorb ) then
         frame.overHealAbsorbGlow:Show();
         myCurrentHealAbsorb = health;
      else
         frame.overHealAbsorbGlow:Hide();
      end
   end
   
   --See how far we're going over the health bar and make sure we don't go too far out of the frame.
   if ( health - myCurrentHealAbsorb + allIncomingHeal > (maxHealth + totalAbsorb) * MAX_INCOMING_HEAL_OVERFLOW ) then
      allIncomingHeal = (maxHealth + totalAbsorb) * MAX_INCOMING_HEAL_OVERFLOW - health + myCurrentHealAbsorb;
   end
   
   local otherIncomingHeal = 0;
   
   --Split up incoming heals.
   if ( allIncomingHeal >= myIncomingHeal ) then
      otherIncomingHeal = allIncomingHeal - myIncomingHeal;
   else
      myIncomingHeal = allIncomingHeal;
   end
   
   local overAbsorb = false;
   
   local healthTexture = frame.healthbar:GetStatusBarTexture();
   local myCurrentHealAbsorbPercent = 0;
   local healAbsorbTexture = nil;
   
   if ( frame.healAbsorbBar ) then
      myCurrentHealAbsorbPercent = myCurrentHealAbsorb / (maxHealth + totalAbsorb);
      
      --If allIncomingHeal is greater than myCurrentHealAbsorb, then the current
      --heal absorb will be completely overlayed by the incoming heals so we don't show it.
      if ( myCurrentHealAbsorb > allIncomingHeal ) then
         local shownHealAbsorb = myCurrentHealAbsorb - allIncomingHeal;
         local shownHealAbsorbPercent = shownHealAbsorb / (maxHealth + totalAbsorb);
         
         healAbsorbTexture = UnitFrameUtil_UpdateFillBar(frame, healthTexture, frame.healAbsorbBar, shownHealAbsorb, -shownHealAbsorbPercent);
         
         --If there are incoming heals the left shadow would be overlayed by the incoming heals
         --so it isn't shown.
         if ( allIncomingHeal > 0 ) then
            frame.healAbsorbBarLeftShadow:Hide();
         else
            frame.healAbsorbBarLeftShadow:SetPoint("TOPLEFT", healAbsorbTexture, "TOPLEFT", 0, 0);
            frame.healAbsorbBarLeftShadow:SetPoint("BOTTOMLEFT", healAbsorbTexture, "BOTTOMLEFT", 0, 0);
            frame.healAbsorbBarLeftShadow:Show();
         end
         
         -- The right shadow is only shown if there are absorbs on the health bar.
         if ( totalAbsorb > 0 ) then
            frame.healAbsorbBarRightShadow:SetPoint("TOPLEFT", healAbsorbTexture, "TOPRIGHT", -8, 0);
            frame.healAbsorbBarRightShadow:SetPoint("BOTTOMLEFT", healAbsorbTexture, "BOTTOMRIGHT", -8, 0);
            frame.healAbsorbBarRightShadow:Show();
         else
            frame.healAbsorbBarRightShadow:Hide();
         end
      else
         frame.healAbsorbBar:Hide();
         frame.healAbsorbBarLeftShadow:Hide();
         frame.healAbsorbBarRightShadow:Hide();
      end
   end
   
   --Show myIncomingHeal on the health bar.
   local incomingHealTexture = UnitFrameUtil_UpdateFillBar(frame, healthTexture, frame.myHealPredictionBar, myIncomingHeal, -myCurrentHealAbsorbPercent);
   
   --Append otherIncomingHeal on the health bar
   if (myIncomingHeal > 0) then
      incomingHealTexture = UnitFrameUtil_UpdateFillBar(frame, incomingHealTexture, frame.otherHealPredictionBar, otherIncomingHeal);
   else
      incomingHealTexture = UnitFrameUtil_UpdateFillBar(frame, healthTexture, frame.otherHealPredictionBar, otherIncomingHeal, -myCurrentHealAbsorbPercent);
   end
   
   --Append absorbs to the correct section of the health bar.
   local appendTexture = nil;
   if ( healAbsorbTexture ) then
      --If there is a healAbsorb part shown, append the absorb to the end of that.
      appendTexture = healAbsorbTexture;
   else
      --Otherwise, append the absorb to the end of the the incomingHeals part;
      appendTexture = incomingHealTexture;
   end
   UnitFrameUtil_UpdateFillBar(frame, appendTexture, frame.totalAbsorbBar, totalAbsorb)
end

local MAX_INCOMING_HEAL_OVERFLOW = 1.00;
function CompactShieldFrame_UpdateHealPrediction(frame)
   local maxHealth = UnitHealthMax(frame.displayedUnit);
   local health = frame.healthBar:GetValue();
   
   if ( maxHealth <= 0 ) then
      return;
   end
   
   if ( not frame.optionTable.displayHealPrediction ) then
      frame.myHealPrediction:Hide();
      frame.otherHealPrediction:Hide();
      frame.totalAbsorb:Hide();
      frame.totalAbsorbOverlay:Hide();
      frame.overAbsorbGlow:Hide();
      frame.myHealAbsorb:Hide();
      frame.myHealAbsorbLeftShadow:Hide();
      frame.myHealAbsorbRightShadow:Hide();
      frame.overHealAbsorbGlow:Hide();
      return;
   end
   
   local myIncomingHeal = UnitGetIncomingHeals(frame.displayedUnit, "player") or 0;
   local allIncomingHeal = UnitGetIncomingHeals(frame.displayedUnit) or 0;
   local totalAbsorb = UnitGetTotalAbsorbs(frame.displayedUnit) or 0;
   local textureScale = (maxHealth + totalAbsorb)/maxHealth;
   
   --We don't fill outside the health bar with healAbsorbs.  Instead, an overHealAbsorbGlow is shown.
   local myCurrentHealAbsorb = UnitGetTotalHealAbsorbs(frame.displayedUnit) or 0;
   if ( health < myCurrentHealAbsorb ) then       
      frame.overHealAbsorbGlow:Show();
      myCurrentHealAbsorb = health;
   else
      frame.overHealAbsorbGlow:Hide();
   end
   
   --See how far we're going over the health bar and make sure we don't go too far out of the frame.
   if ( health - myCurrentHealAbsorb + allIncomingHeal > (maxHealth + totalAbsorb) * MAX_INCOMING_HEAL_OVERFLOW ) then
      allIncomingHeal = (maxHealth + totalAbsorb) * MAX_INCOMING_HEAL_OVERFLOW - health + myCurrentHealAbsorb;
   end
   
   local otherIncomingHeal = 0;
   
   --Split up incoming heals.
   if ( allIncomingHeal >= myIncomingHeal ) then
      otherIncomingHeal = allIncomingHeal - myIncomingHeal;
   else
      myIncomingHeal = allIncomingHeal;
   end
   
   local overAbsorb = false;
   
   local healthTexture = frame.healthBar:GetStatusBarTexture();
   local myCurrentHealAbsorbPercent = myCurrentHealAbsorb / (maxHealth + totalAbsorb);
   local healAbsorbTexture = nil;
   
   --If allIncomingHeal is greater than myCurrentHealAbsorb, then the current
   --heal absorb will be completely overlayed by the incoming heals so we don't show it.
   if ( myCurrentHealAbsorb > allIncomingHeal ) then
      local shownHealAbsorb = myCurrentHealAbsorb - allIncomingHeal;
      local shownHealAbsorbPercent = shownHealAbsorb / (maxHealth + totalAbsorb);
      healAbsorbTexture = CompactUnitFrameUtil_UpdateFillBar(frame, healthTexture, frame.myHealAbsorb, shownHealAbsorb, -shownHealAbsorbPercent);
      
      --If there are incoming heals the left shadow would be overlayed by the incoming heals
      --so it isn't shown.
      if ( allIncomingHeal > 0 ) then
         frame.myHealAbsorbLeftShadow:Hide();
      else
         frame.myHealAbsorbLeftShadow:SetPoint("TOPLEFT", healAbsorbTexture, "TOPLEFT", 0, 0);
         frame.myHealAbsorbLeftShadow:SetPoint("BOTTOMLEFT", healAbsorbTexture, "BOTTOMLEFT", 0, 0);
         frame.myHealAbsorbLeftShadow:Show();
      end
      
      -- The right shadow is only shown if there are absorbs on the health bar.
      if ( totalAbsorb > 0 ) then
         frame.myHealAbsorbRightShadow:SetPoint("TOPLEFT", healAbsorbTexture, "TOPRIGHT", -8, 0);
         frame.myHealAbsorbRightShadow:SetPoint("BOTTOMLEFT", healAbsorbTexture, "BOTTOMRIGHT", -8, 0);
         frame.myHealAbsorbRightShadow:Show();
      else
         frame.myHealAbsorbRightShadow:Hide();
      end
   else
      frame.myHealAbsorb:Hide();
      frame.myHealAbsorbRightShadow:Hide();
      frame.myHealAbsorbLeftShadow:Hide();
   end
   
   --Show myIncomingHeal on the health bar.
   local incomingHealsTexture = CompactUnitFrameUtil_UpdateFillBar(frame, healthTexture, frame.myHealPrediction, myIncomingHeal, -myCurrentHealAbsorbPercent);
   --Append otherIncomingHeal on the health bar.
   incomingHealsTexture = CompactUnitFrameUtil_UpdateFillBar(frame, incomingHealsTexture, frame.otherHealPrediction, otherIncomingHeal);
   
   --Appen absorbs to the correct section of the health bar.
   local appendTexture = nil;
   if ( healAbsorbTexture ) then
      --If there is a healAbsorb part shown, append the absorb to the end of that.
      appendTexture = healAbsorbTexture;
   else
      --Otherwise, append the absorb to the end of the the incomingHeals part;
      appendTexture = incomingHealsTexture;
   end
   CompactUnitFrameUtil_UpdateFillBar(frame, appendTexture, frame.totalAbsorb, totalAbsorb)
end

function ShieldStatusBar_UpdateTextString(textStatusBar, maxHealth)
   local textString = textStatusBar.TextString;
   if(textString) then
      local value = textStatusBar:GetValue();
      local valueMin, valueMax = textStatusBar:GetMinMaxValues();
      TextStatusBar_UpdateTextStringWithValues(textStatusBar, textString, value, valueMin, maxHealth);
   end
end
